(ns puppetlabs.redrover.hierarchy
  "This namespace contains functions for building a 3.7+ node group hierarchy
  out of of old 3.3 dashboard classification groups. Throughout the docstrings
  and argument vectors in this namespace, the new 3.7+ node groups are called
  `node groups`, the in-memory storage instance that contains them is called
  a `node group hierarchy` or simply `hierarchy`, and the old 3.3 dashboard
  classification groups are called `old groups`."
  (:require [clojure.set :as set]
            [schema.core :as sc]
            [puppetlabs.redrover.schema :refer [OldGroup OldNode]]
            [puppetlabs.classifier.schema :refer [Group RuleCondition]]
            [puppetlabs.classifier.storage :as storage]
            [puppetlabs.classifier.storage.memory :refer [in-memory-storage]]
            [puppetlabs.classifier.storage.naive :as naive])
  (:import java.util.UUID))

;;
;; Searching Node Group Hierarchies
;;

(defn find-group-named
  [hierarchy group-name]
  (->> (storage/get-groups hierarchy)
    (filter #(= (:name %) group-name))
    first))

(defn find-parent
  "Given a node group hierarchy and an old group, attempt to find by name the
  node group in the hierarchy that's the first parent of the old group. If this
  parent node group is found, return it; if not, then return nil.  "
  [hierarchy old-group]
  (if-let [parent-name (first (:parents old-group))]
    (find-group-named hierarchy parent-name)))

;;
;; Problematic Old Group & Node Tests
;;

(defn multiple-parents?
  "Takes an old group. If the old group has multiple parents, returns it;
  otherwise, returns nil."
  [old-group]
  (if (> (count (:parents old-group)) 1)
    old-group))

(defn missing-parent?
  "Determines whether the first parent of old-group is missing in the given node
  group hierarchy. If so, returns the group; otherwise, returns nil. If the
  group has no parent, it is not considered to have a missing parent. If given
  a group with more than one parent, throws an AssertionError."
  [hierarchy old-group]
  {:pre [(not (multiple-parents? old-group))]}
  (let [{:keys [parents]} old-group]
    (if (and (pos? (count parents))
             (not (find-parent hierarchy old-group)))
      old-group)))

(declare pinned-nodes-rule->nodes)

(defn ancestral-nodes-mismatch?
  "Determine whether old-group matches any nodes that at least one of its
  ancestors doesn't match. If this is the case, return old-group; otherwise,
  return nil. Throws an AssertionError if old-group's first parent is defined
  and cannot be found in the hierarchy, or if old-group has multiple parents."
  [hierarchy old-group]
  {:pre [(not (missing-parent? hierarchy old-group))]}
  (if (empty? (:parents old-group))
    nil
    (let [parent (find-parent hierarchy old-group)
          ancestors (concat [parent] (naive/get-ancestors hierarchy parent))
          group-nodes (-> (:nodes old-group) set)
          nodes-mismatch? (fn [{:keys [rule id]}]
                            (and (not= id storage/root-group-uuid)
                                 (let [ancestor-nodes (-> (pinned-nodes-rule->nodes rule) set)]
                                   (not (set/subset? group-nodes ancestor-nodes)))))]
      (if (some nodes-mismatch? ancestors)
        old-group))))

(defn has-classification?
  "Takes an old node or group, and returns it if it has a non-empty
  classification (i.e., at least one class or variable is set). Otherwise,
  returns nil."
  [old-node-or-group]
  (let [{:keys [classes variables]} old-node-or-group]
    (if (or (seq classes) (seq variables))
      old-node-or-group)))

(defn- matches-pinned-node?
  [node-name group]
  (some #{node-name} (pinned-nodes-rule->nodes (:rule group))))

(defn node-in-group?
  "Takes a node group hierarchy and an old node, and returns the node if it's
  a pinned node of any group in the hierarchy. The single node name in the
  :nodes field of the old-node is used when searching for the node in groups.
  If the node returns nil."
  [hierarchy old-node]
  (let [node-name (-> (:nodes old-node) first)
        groups (->> (storage/get-groups hierarchy)
                 (remove #(= (:id %) storage/root-group-uuid)))]
    (if (some (partial matches-pinned-node? node-name) groups)
      old-node)))

;;
;; Transforming Old Groups to Node Groups
;;

(sc/defn pinned-nodes-rule->nodes :- [sc/Str]
  [rule :- RuleCondition]
  (let [equality-clauses (rest rule)]
    (map #(nth % 2) equality-clauses)))

(sc/defn nodes->pinned-nodes-rule :- RuleCondition
  [nodes]
  (apply vector "or" (for [node nodes]
                       ["=" "name" node])))

(sc/defn old-group->new-group :- Group
  "Transform an old group into a node group.

  The node group's environment will always be \"production\", the
  environment-trumps flag will always be unset, and the parent will always be
  the root group.

  The rule will be one large disjunction over the names of the nodes of the 3.3
  group.

  The id will be a randomly generated UUID.

  All the other keys (name, description, classes, and variables) will be carried
  over unmodified."
  [old-group]
  (let [{:keys [classes description name nodes parents variables]} old-group
        group {:id (UUID/randomUUID)
               :name name
               :description (or description "")
               :parent storage/root-group-uuid
               :environment "production"
               :environment-trumps false
               :classes classes
               :variables variables}]
    (if (empty? nodes)
      group
      (assoc group :rule (nodes->pinned-nodes-rule nodes)))))

;;
;; Creating & Building a Node Group Hierarchy
;;

(defn new-hierarchy
  "Returns a fresh node group hierarchy that contains only the root group."
  []
  (in-memory-storage {}))

(defn add-old-group!
  "Add a single old group to the given node group hierarchy. Assumes that the
  old group has at most one parent, and that if it has a parent, that parent is
  already in the node group hierarchy. Throws an AssertionError if either of
  these assumptions are violated.

  Since storage is side-effectful, the return value is always `nil`."
  [hierarchy old-group]
  {:pre [(not (missing-parent? hierarchy old-group))]}
  (let [{:keys [name parents]} old-group
        parent-group (find-parent hierarchy old-group)
        new-group (merge (old-group->new-group old-group)
                         (if-let [parent-id (:id parent-group)]
                           {:parent parent-id}))]
    (storage/create-group hierarchy new-group)
    nil))
